from psyneulink.core import llvm as pnlvm
from psyneulink.library.compositions.pytorchllvmhelper import *

__all__ = ['AdamOptimizer', 'SGDOptimizer']

class Optimizer():

    def __init__(self, pytorch_model):
        self._pytorch_model = pytorch_model
        self._composition = pytorch_model._composition

        self._DELTA_W_NUM = 0

    # gets the type of the delta_w struct
    def _get_delta_w_struct_type(self, ctx):
        delta_w = [None] * len(self._pytorch_model.projections)
        for idx, proj in enumerate(self._pytorch_model.projections):
            proj_matrix = proj.matrix
            dim_x, dim_y = proj_matrix.shape

            matrix = pnlvm.ir.ArrayType(
                pnlvm.ir.ArrayType(
                    ctx.float_ty,
                    dim_y
                ),
                dim_x
            )
            delta_w[idx] = matrix

        delta_w = pnlvm.ir.types.LiteralStructType(delta_w)
        return delta_w

    def _get_optimizer_struct_type(self, ctx, extra_types=[]):
        structs = (self._get_delta_w_struct_type(ctx), *extra_types)
        return pnlvm.ir.types.LiteralStructType(structs)

    # inserts logic that zeroes out a gradient struct
    def _gen_zero_gradient_struct(self, ctx, builder, grad_struct):
        builder.store(grad_struct.type.pointee(None),grad_struct)

    def zero_grad(self, ctx):
        name = self._composition.name + "_ZERO_GRAD"

        args = [self._get_optimizer_struct_type(ctx).as_pointer()]
        builder = ctx.create_llvm_function(args, self, name)
        llvm_func = builder.function
        optim_struct = llvm_func.args[0]

        delta_w_struct = builder.gep(
            optim_struct, [ctx.int32_ty(0), ctx.int32_ty(self._DELTA_W_NUM)])
        self._gen_zero_gradient_struct(ctx, builder, delta_w_struct)
        builder.ret_void()

        return llvm_func

    def _gen_llvm_function(self, *, ctx:pnlvm.LLVMBuilderContext, tags:frozenset):
        return self.step(ctx)

    # to be implemented by child classes - steps the optimizer
    def step(self, ctx):
        raise Exception("Unimplemented method!")

# Class that is used to represent a compiled optimizer - aims to reimplement the logic of torch.optimizer in the form of llvmlite compileable code


class AdamOptimizer(Optimizer):
    """Implements compiled ADAM Optimizer ( from paper https://arxiv.org/pdf/1412.6980.pdf  )"""
    # sets up parameters of model & the information required for forward computation
    def __init__(self, pytorch_model, lr=1e-3, betas=(.9, .999), eps=1e-8, weight_decay=0,):
        super().__init__(pytorch_model)
        self.lr = lr
        self.betas = betas
        self.eps = eps
        self.weight_decay = weight_decay

        self._M_T_NUM = 1
        self._V_T_NUM = 2
        self._T_NUM = 3

    def _get_optimizer_struct_type(self, ctx):
        m_t = self._get_delta_w_struct_type(ctx)  # current timestep first moment
        v_t = self._get_delta_w_struct_type(ctx)  # current timestep second moment
        time_counter = ctx.float_ty  # keeps track of timestep

        extra_types = [m_t, v_t, time_counter]
        return super()._get_optimizer_struct_type(ctx, extra_types=extra_types)

    # steps the adam optimizer (methodology: https://arxiv.org/pdf/1412.6980.pdf )
    def step(self, ctx):
        name = self._composition.name + "_ADAM_STEP"

        args = [self._get_optimizer_struct_type(ctx).as_pointer(),
                ctx.get_param_struct_type(self._composition).as_pointer()]
        builder = ctx.create_llvm_function(args, self, name)
        llvm_func = builder.function
        optim_struct, params = llvm_func.args

        # setup values
        zero = ctx.int32_ty(0)
        one_float = ctx.float_ty(1)

        delta_w = builder.gep(optim_struct, [zero, ctx.int32_ty(self._DELTA_W_NUM)])
        m_t = builder.gep(optim_struct, [zero, ctx.int32_ty(self._M_T_NUM)])
        v_t = builder.gep(optim_struct, [zero, ctx.int32_ty(self._V_T_NUM)])
        t = builder.gep(optim_struct, [zero, ctx.int32_ty(self._T_NUM)])

        # get methods needed
        pow = ctx.import_llvm_function("__pnl_builtin_pow")
        sqrt = ctx.get_builtin("sqrt", [ctx.float_ty])

        lr = ctx.float_ty(self.lr)
        eps = ctx.float_ty(self.eps)
        b1 = ctx.float_ty(self.betas[0])
        b2 = ctx.float_ty(self.betas[1])
        one_minus_b1 = builder.fsub(one_float, b1)
        one_minus_b2 = builder.fsub(one_float, b2)

        # 1) increment t
        builder.store(builder.fadd(builder.load(t), one_float), t)
        t_val = builder.load(t)
        # 1.5) calculate values to be used later (based on incremented t)
        b1_pow = builder.call(pow, [b1, t_val])
        b2_pow = builder.call(pow, [b2, t_val])
        one_minus_b1_pow = builder.fsub(one_float, b1_pow)
        one_minus_b2_pow = builder.fsub(one_float, b2_pow)

        pnlvm.helpers.printf(
                builder, f"%f b1_pow_sub %f\nb2 pow sub %f\n",t_val, one_minus_b1_pow, one_minus_b2_pow)

        # 2) update first moments
        for idx, proj in enumerate(self._pytorch_model.projections):
            proj_idx_ir = ctx.int32_ty(idx)

            m_t_ptr = builder.gep(m_t, [zero, proj_idx_ir])
            delta_w_ptr = builder.gep(delta_w, [zero, proj_idx_ir])

            # m_t = m_t * b1
            gen_inject_mat_scalar_mult(ctx, builder, m_t_ptr, b1, m_t_ptr)

            # (1 - b1)*g_t
            tmp_val = gen_inject_mat_scalar_mult(ctx, builder, delta_w_ptr, one_minus_b1)

            # m_t = m_t + (1-b1)*g_t
            gen_inject_mat_add(ctx, builder, m_t_ptr, tmp_val, m_t_ptr)

            pnlvm.helpers.printf_float_matrix(builder, m_t_ptr, prefix=f"mt val: {proj.sender._mechanism} -> {proj.receiver._mechanism}\n", override_debug=False)
        # 3) update second moments
        for idx, proj in enumerate(self._pytorch_model.projections):
            proj_idx_ir = ctx.int32_ty(idx)

            v_t_ptr = builder.gep(v_t, [zero, proj_idx_ir])
            delta_w_ptr = builder.gep(delta_w, [zero, proj_idx_ir])

            # v_t = v_t * b2
            gen_inject_mat_scalar_mult(ctx, builder, v_t_ptr, b2, v_t_ptr)

            # g_t * g_t
            delta_w_sqrd = gen_inject_mat_hadamard(ctx, builder, delta_w_ptr, delta_w_ptr)

            # (1-b2)*(g_t)^2
            gen_inject_mat_scalar_mult(ctx, builder, delta_w_sqrd, one_minus_b2, delta_w_sqrd)

            # v_t = v_t + (1-b2)*(g_t)^2
            gen_inject_mat_add(ctx, builder, v_t_ptr, delta_w_sqrd, v_t_ptr)

        # 4) update weights
        # this is the new learning rate to use
        # NOTE: This differs from the version specified in the paper to numerically match with pytorch's implementation
        step_size = builder.fdiv(lr, one_minus_b1_pow)
        step_size = pnlvm.helpers.fneg(builder, step_size)

        for idx, proj in enumerate(self._pytorch_model.projections):
            proj_idx_ir = ctx.int32_ty(idx)

            m_t_ptr = builder.gep(
                m_t, [zero, proj_idx_ir])
            v_t_ptr = builder.gep(
                v_t, [zero, proj_idx_ir])
            delta_w_ptr = builder.gep(
                delta_w, [zero, proj_idx_ir])

            pnlvm.helpers.printf_float_matrix(builder, delta_w_ptr, prefix=f"grad val: {proj.sender._mechanism} -> {proj.receiver._mechanism}\n", override_debug=False)

            # this is messy - #TODO - cleanup this
            weights_llvmlite = proj._extract_llvm_matrix(ctx, builder, params)
            dim_x, dim_y = proj.matrix.shape

            weight_row = None
            pnlvm.helpers.printf(builder, "biascorr2 %.20f\n", one_minus_b2_pow, override_debug=False)
            with pnlvm.helpers.for_loop_zero_inc(builder, ctx.int32_ty(dim_x), "optimizer_w_upd_outer") as (b1, weight_row):
                weight_column = None
                with pnlvm.helpers.for_loop_zero_inc(b1, ctx.int32_ty(dim_y), "optimizer_w_upd_inner") as (b2, weight_column):
                    # sqrt(v_t) + eps
                    v_t_value = b2.load(b2.gep(v_t_ptr, [zero, weight_row, weight_column]))
                    value = b2.call(sqrt, [v_t_value])
                    denom = b2.call(sqrt, [one_minus_b2_pow])
                    value = b2.fdiv(value, denom)
                    value = b2.fadd(value, eps)
                    pnlvm.helpers.printf(builder, "val %.20f\n", value, override_debug=False)
                    # alpha_t * m_t
                    m_t_value = b2.load(b2.gep(
                        m_t_ptr, [zero, weight_row, weight_column]))

                    # value = alpha_t * m_t / (sqrt(v_t) + eps)
                    value = b2.fdiv(m_t_value, value)
                    value = b2.fmul(step_size, value)

                    old_weight_ptr = b2.gep(
                        weights_llvmlite, [zero, weight_row, weight_column])

                    # new_weight = old_weight - value
                    value = b2.fadd(b2.load(old_weight_ptr), value)
                    b2.store(value, old_weight_ptr)

                pnlvm.helpers.printf(b1, "\n", override_debug=False)

        pnlvm.helpers.printf(builder, f"\t\t\tOPTIM DONE UPDATE\n",override_debug=False)

        builder.ret_void()

        return llvm_func


class SGDOptimizer(Optimizer):
    """Implements compiled Stocastic Gradient Descent optimizer (without momentum)"""
    # sets up parameters of model & the information required for forward computation
    def __init__(self, pytorch_model, lr=1e-3):
        super().__init__(pytorch_model)
        self.lr = lr

    # steps the sgd optimizer (methodology: https://arxiv.org/pdf/1412.6980.pdf )
    def step(self, ctx):
        name = self._composition.name + "_SGD_STEP"

        args = [self._get_optimizer_struct_type(ctx).as_pointer(),
                ctx.get_param_struct_type(self._composition).as_pointer()]
        builder = ctx.create_llvm_function(args, self, name)
        llvm_func = builder.function
        optim_struct, params = llvm_func.args

        zero = ctx.int32_ty(0)
        delta_w = builder.gep(optim_struct, [zero, ctx.int32_ty(self._DELTA_W_NUM)])

        lr = ctx.float_ty(self.lr)

        # update weights
        for idx, proj in enumerate(self._pytorch_model.projections):
            delta_w_ptr = builder.gep(delta_w, [zero, ctx.int32_ty(idx)])
            weights_llvmlite = proj._extract_llvm_matrix(ctx, builder, params)

            multiplied_delta_w = gen_inject_mat_scalar_mult(ctx, builder, delta_w_ptr, lr)
            gen_inject_mat_sub(ctx, builder, weights_llvmlite, multiplied_delta_w, weights_llvmlite)

        builder.ret_void()

        return llvm_func




import numpy as np

import psyneulink.core.llvm as pnlvm
import psyneulink.core.components.functions.function as Function
import psyneulink.core.components.functions.objectivefunctions as Functions
import psyneulink.core.components.functions.transferfunctions
import psyneulink.core.globals.keywords as kw
import pytest

SIZE=10
# Some metrics (CROSS_ENTROPY) don't like 0s
test_var = np.random.rand(SIZE) + Function.EPSILON
hollow_matrix = Function.get_matrix(kw.HOLLOW_MATRIX, SIZE, SIZE)
v1 = test_var
v2 = np.dot(hollow_matrix * hollow_matrix, v1)
norm = len(v1)

test_data = [
    (kw.ENTROPY, False, -np.sum(v1 * np.log(v2))),
    (kw.ENTROPY, True, -np.sum(v1 * np.log(v2)) / norm),
    (kw.ENERGY, False, -np.sum(v1 * v2) / 2),
    (kw.ENERGY, True, (-np.sum(v1 * v2) / 2) / norm**2),
]

# use list, naming function produces ugly names
names = [
    "ENTROPY",
    "ENTROPY NORMALIZED",
    "ENERGY",
    "ENERGY NORMALIZED",
]

@pytest.mark.function
@pytest.mark.stability_function
@pytest.mark.benchmark
@pytest.mark.parametrize("metric, normalize, expected", test_data, ids=names)
@pytest.mark.parametrize("variable", [test_var, test_var.astype(np.float32)], ids=["float", "float32"] )
@pytest.mark.parametrize('mode', ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_basic(variable, metric, normalize, expected, benchmark, mode):
    f = Functions.Stability(default_variable=variable, metric=metric, normalize=normalize)
    if mode == 'Python':
        EX = f.function
    elif mode == 'LLVM':
        e = pnlvm.execution.FuncExecution(f)
        EX = e.execute
    elif mode == 'PTX':
        e = pnlvm.execution.FuncExecution(f)
        EX = e.cuda_execute

    benchmark.group = "DistanceFunction " + metric + ("-normalized" if normalize else "")
    res = benchmark(EX, variable)
    assert np.allclose(res, expected)
    assert np.isscalar(res) or len(res) == 1



import numpy as np
import psyneulink as pnl
import psyneulink.core.components.functions.transferfunctions

from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import AccumulatorIntegrator
from psyneulink.core.components.functions.transferfunctions import Logistic
from psyneulink.core.components.mechanisms.modulatory.control.gating.gatingmechanism import GatingMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.globals.keywords import \
    DEFAULT_VARIABLE, FUNCTION, FUNCTION_PARAMS, INITIALIZER, RATE, TARGET_MECHANISM, VALUE
from psyneulink.core.compositions.composition import Composition

def test_gating_with_composition():
    """Tests same configuration as control of InputPort in tests/mechansims/test_identicalness_of_control_and_gating
    """
    Input_Layer = TransferMechanism(name='Input Layer', function=Logistic, size=2)
    Hidden_Layer_1 = TransferMechanism(name='Hidden Layer_1', function=Logistic, size=5)
    Hidden_Layer_2 = TransferMechanism(name='Hidden Layer_2', function=Logistic, size=4)
    Output_Layer = TransferMechanism(name='Output Layer', function=Logistic, size=3)

    Gating_Mechanism = GatingMechanism(size=[1], gate=[Hidden_Layer_1, Hidden_Layer_2, Output_Layer])

    Input_Weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
    Middle_Weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
    Output_Weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)
    # This projection is specified in add_backpropagation_learning_pathway method below
    Input_Weights = MappingProjection(name='Input Weights',matrix=Input_Weights_matrix)
    # This projection is "discovered" by add_backpropagation_learning_pathway method below
    Middle_Weights = MappingProjection(name='Middle Weights',sender=Hidden_Layer_1,receiver=Hidden_Layer_2,
        matrix={
            VALUE: Middle_Weights_matrix,
            FUNCTION: AccumulatorIntegrator,
            FUNCTION_PARAMS: {
                DEFAULT_VARIABLE: Middle_Weights_matrix,
                INITIALIZER: Middle_Weights_matrix,
                RATE: Middle_Weights_matrix
            },
        }
    )
    Output_Weights = MappingProjection(sender=Hidden_Layer_2, receiver=Output_Layer, matrix=Output_Weights_matrix)

    pathway = [Input_Layer, Input_Weights, Hidden_Layer_1, Hidden_Layer_2, Output_Layer]
    comp = Composition()
    backprop_pathway = comp.add_backpropagation_learning_pathway(pathway=pathway,
                                                                 loss_function=None)
    # c.add_linear_processing_pathway(pathway=z)
    comp.add_node(Gating_Mechanism)

    stim_list = {
        Input_Layer: [[-1, 30]],
        Gating_Mechanism: [1.0],
        backprop_pathway.target: [[0, 0, 1]]}

    comp.learn(num_trials=3, inputs=stim_list)

    expected_results = [[[0.81493513, 0.85129046, 0.88154205]],
                        [[0.81331773, 0.85008207, 0.88157851]],
                        [[0.81168332, 0.84886047, 0.88161468]]]

    assert np.allclose(comp.results, expected_results)

    stim_list[Gating_Mechanism]=[0.0]
    results = comp.learn(num_trials=1, inputs=stim_list)
    expected_results = [[[0.5, 0.5, 0.5]]]
    assert np.allclose(results, expected_results)

    stim_list[Gating_Mechanism]=[2.0]
    results = comp.learn(num_trials=1, inputs=stim_list)
    expected_results = [[0.96941429, 0.9837254 , 0.99217549]]
    assert np.allclose(results, expected_results)

def test_gating_with_UDF_with_composition():
    def my_linear_fct(
        x,
        m=2.0,
        b=0.0,
        params={
            pnl.ADDITIVE_PARAM: 'b',
            pnl.MULTIPLICATIVE_PARAM: 'm'
        }
    ):
        return m * x + b

    def my_simple_linear_fct(
        x,
        m=1.0,
        b=0.0
    ):
        return m * x + b

    def my_exp_fct(
        x,
        r=1.0,
        # b=pnl.CONTROL,
        b=0.0,
        params={
            pnl.ADDITIVE_PARAM: 'b',
            pnl.MULTIPLICATIVE_PARAM: 'r'
        }
    ):
        return x**r + b

    def my_sinusoidal_fct(
        input,
        phase=0,
        amplitude=1,
        params={
            pnl.ADDITIVE_PARAM: 'phase',
            pnl.MULTIPLICATIVE_PARAM: 'amplitude'
        }
    ):
        frequency = input[0]
        t = input[1]
        return amplitude * np.sin(2 * np.pi * frequency * t + phase)

    Input_Layer = pnl.TransferMechanism(
        name='Input_Layer',
        default_variable=np.zeros((2,)),
        function=psyneulink.core.components.functions.transferfunctions.Logistic
    )

    Output_Layer = pnl.TransferMechanism(
        name='Output_Layer',
        default_variable=[0, 0, 0],
        function=psyneulink.core.components.functions.transferfunctions.Linear,
        # function=pnl.Logistic,
        # output_ports={pnl.NAME: 'RESULTS USING UDF',
        #                pnl.VARIABLE: [(pnl.OWNER_VALUE,0), pnl.TIME_STEP],
        #                pnl.FUNCTION: my_sinusoidal_fct}
        output_ports={
            pnl.NAME: 'RESULTS USING UDF',
            # pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
            pnl.FUNCTION: psyneulink.core.components.functions.transferfunctions.Linear(slope=pnl.GATING)
        }
    )

    Gating_Mechanism = pnl.GatingMechanism(
        size=[1],
        gating_signals=[
            # Output_Layer
            Output_Layer.output_port,
        ]
    )

    comp = Composition()
    comp.add_linear_processing_pathway(pathway=[Input_Layer, Output_Layer])
    comp.add_node(Gating_Mechanism)

    stim_list = {
        Input_Layer: [[-1, 30], [-1, 30], [-1, 30], [-1, 30]],
        Gating_Mechanism: [[0.0], [0.5], [1.0], [2.0]]
    }

    comp.run(num_trials=4, inputs=stim_list)

    expected_results = [
        [np.array([0., 0., 0.])],
        [np.array([0.63447071, 0.63447071, 0.63447071])],
        [np.array([1.26894142, 1.26894142, 1.26894142])],
        [np.array([2.53788284, 2.53788284, 2.53788284])]
    ]

    np.testing.assert_allclose(comp.results, expected_results)



from psyneulink.core.compositions.composition import Composition
from psyneulink.core.components.functions.transferfunctions import Exponential
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.mechanisms.modulatory.control.gating.gatingmechanism import GatingMechanism
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.modulatorysignals.gatingsignal import GatingSignal

from psyneulink.core.globals.keywords import SLOPE, RESULT

class TestControlSignals:
    def test_control_signal_intensity_cost_function(self):

        mech = TransferMechanism()
        ctl_sig = ControlSignal(projections=[(SLOPE, mech)],
                                intensity_cost_function=Exponential(rate=1))
        ctl_mech = ControlMechanism(control_signals=[ctl_sig])
        ctl_mech.execute()
        assert True

    def test_alias_equivalence_for_modulates_and_projections(self):
        inputs = [1, 9, 4, 3, 2]
        comp1 = Composition()
        tMech1 = TransferMechanism()
        tMech2 = TransferMechanism()
        cMech1 = ControlMechanism(control_signals=ControlSignal(modulates=(SLOPE, tMech2)),
                                  objective_mechanism=ObjectiveMechanism(monitor=(RESULT, tMech2)))
        comp1.add_nodes([tMech1, tMech2, cMech1])
        comp1.add_linear_processing_pathway([cMech1, tMech1, tMech2])
        comp1.run(inputs=inputs)
        comp2 = Composition()
        tMech3 = TransferMechanism()
        tMech4 = TransferMechanism()
        cMech2 = ControlMechanism(control_signals=ControlSignal(projections=(SLOPE, tMech4)),
                                  objective_mechanism=ObjectiveMechanism(monitor=(RESULT, tMech4)))
        comp2.add_nodes([tMech3, tMech4, cMech2])
        comp2.add_linear_processing_pathway([cMech2, tMech3, tMech4])
        comp2.run(inputs=inputs)
        assert comp1.results == comp2.results

class TestGatingSignals:
    def test_alias_equivalence_for_modulates_and_projections(self):
        inputs = [1123, 941, 43, 311, 21]
        Tx1 = TransferMechanism()
        Ty1 = TransferMechanism()
        G1 = GatingMechanism(gating_signals=[GatingSignal(modulates=Tx1)])
        comp1 = Composition()
        comp1.add_nodes([Tx1, Ty1, G1])
        comp1.add_linear_processing_pathway([Tx1, Ty1, G1, Tx1])
        comp1.run(inputs={Tx1: inputs})

        Tx2 = TransferMechanism()
        Ty2 = TransferMechanism()
        G2 = GatingMechanism(gating_signals=[GatingSignal(projections=Tx2)])
        comp2 = Composition()
        comp2.add_nodes([Tx2, Ty2, G2])
        comp2.add_linear_processing_pathway([Tx2, Ty2, G2, Tx2])
        comp2.run(inputs={Tx2: inputs})
        assert comp1.results == comp2.results



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  CompositionInterfaceMechanism *************************************************

"""

Contents
--------

  * `CompositionInterfaceMechanism_Overview`
  * `CompositionInterfaceMechanism_Creation`
  * `CompositionInterfaceMechanism_Structure`
  * `CompositionInterfaceMechanism_Execution`
  * `CompositionInterfaceMechanism_Class_Reference`


.. _CompositionInterfaceMechanism_Overview:

Overview
--------

A CompositionInterfaceMechanism stores inputs from outside the Composition so that those can be delivered to the
Composition's `INPUT <NodeRole.INPUT>` Mechanism(s).

.. _CompositionInterfaceMechanism_Creation:

Creating an CompositionInterfaceMechanism
-----------------------------------------

A CompositionInterfaceMechanism is created automatically when an `INPUT <NodeRole.INPUT>` Mechanism is identified in a
Composition. When created, the CompositionInterfaceMechanism's OutputPort is set directly by the Composition. This
Mechanism should never be executed, and should never be created by a user.

.. _CompositionInterfaceMechanism_Structure

Structure
---------

[TBD]

.. _CompositionInterfaceMechanism_Execution

Execution
---------

[TBD]

.. _CompositionInterfaceMechanism_Class_Reference:

Class Reference
---------------

"""

import warnings
import typecheck as tc

from collections.abc import Iterable

from psyneulink.core.components.functions.transferfunctions import Identity
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import COMPOSITION_INTERFACE_MECHANISM, INPUT_PORTS, OUTPUT_PORTS, PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = ['CompositionInterfaceMechanism']


class CompositionInterfaceMechanism(ProcessingMechanism_Base):
    """
    CompositionInterfaceMechanism(  \
        function=Identity())

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that acts as interface between a Composition and its
    inputs from and outputs to the environment or other Mechanisms (if it is a nested Composition).

    See `Mechanism <Mechanism_Class_Reference>` for arguments and additonal attributes.

    Attributes
    ----------

    function : InterfaceFunction : default Identity
        the function used to transform the variable before assigning it to the Mechanism's OutputPort(s)

    """

    componentType = COMPOSITION_INTERFACE_MECHANISM
    outputPortTypes = [OutputPort, ControlSignal]

    classPreferenceLevel = PreferenceLevel.TYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'CompositionInterfaceMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <CompositionInterfaceMechanism.function>`

                    :default value: `Identity`
                    :type: `Function`
        """
        function = Parameter(Identity, stateful=False, loggable=False)

    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 size=None,
                 input_ports: tc.optional(tc.optional(tc.any(Iterable, Mechanism, OutputPort, InputPort))) = None,
                 function=None,
                 composition=None,
                 port_map=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None):

        if default_variable is None and size is None:
            default_variable = self.class_defaults.variable
        self.composition = composition
        self.port_map = port_map
        self.connected_to_composition = False
        self.user_added_ports = {
            INPUT_PORTS: set(),
            OUTPUT_PORTS: set()
        }
        super(CompositionInterfaceMechanism, self).__init__(default_variable=default_variable,
                                                            size=size,
                                                            input_ports=input_ports,
                                                            function=function,
                                                            params=params,
                                                            name=name,
                                                            prefs=prefs,
                                                            )

    @handle_external_context()
    def add_ports(self, ports, context=None):
        ports = super(CompositionInterfaceMechanism, self).add_ports(ports, context=context)
        if context.source == ContextFlags.COMMAND_LINE:
            warnings.warn(
                'You are attempting to add custom ports to a CIM, which can result in unpredictable behavior and '
                'is therefore recommended against. If suitable, you should instead add ports to the mechanism(s) '
                'that project to or are projected to from the CIM.')
            if ports[INPUT_PORTS]:
                self.user_added_ports[INPUT_PORTS].update([port for port in ports[INPUT_PORTS].data])
            if ports[OUTPUT_PORTS]:
                self.user_added_ports[OUTPUT_PORTS].update([port for port in ports[OUTPUT_PORTS].data])
        return ports

    @handle_external_context()
    def remove_ports(self, ports, context=None):
        super(CompositionInterfaceMechanism, self).remove_ports(ports, context)
        input_ports_marked_for_deletion = set()
        for port in self.user_added_ports[INPUT_PORTS]:
            if port not in self.input_ports:
                input_ports_marked_for_deletion.add(port)
        self.user_added_ports[INPUT_PORTS] = self.user_added_ports[INPUT_PORTS] - input_ports_marked_for_deletion
        output_ports_marked_for_deletion = set()
        for port in self.user_added_ports[OUTPUT_PORTS]:
            if port not in self.output_ports:
                output_ports_marked_for_deletion.add(port)
        self.user_added_ports[OUTPUT_PORTS] = self.user_added_ports[OUTPUT_PORTS] - output_ports_marked_for_deletion



import numpy as np
from psyneulink import *
from gym_forager.envs.forager_env import ForagerEnv

# Runtime Switches:
RENDER = True
PERCEPT_DISTORT = False
PNL_COMPILE = False

# *********************************************************************************************************************
# *********************************************** CONSTANTS ***********************************************************
# *********************************************************************************************************************

# These should probably be replaced by reference to ForagerEnv constants:
obs_len = 2
action_len = 2
player_coord_idx = slice(0, 2)
predator_coord_idx = slice(3, 5)
prey_coord_idx = slice(6, 8)
player_value_idx = 2
predator_value_idx = 5
prey_value_idx = 8

player_len = prey_len = predator_len = obs_len

# *********************************************************************************************************************
# **************************************  MECHANISMS AND COMPOSITION  *************************************************
# *********************************************************************************************************************

if PERCEPT_DISTORT:
    player = ProcessingMechanism(size=prey_len, function=GaussianDistort(variance=0), name="PLAYER OBS")
    prey = ProcessingMechanism(size=prey_len, function=GaussianDistort(variance=0), name="PREY OBS")
else:
    player = TransferMechanism(size=prey_len, name="PLAYER OBS")
    prey = TransferMechanism(size=prey_len, name="PREY OBS")

# For future use:
values = TransferMechanism(size=3, name="AGENT VALUES")
reward = TransferMechanism(name="REWARD")

# Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
# note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT', sample=player, target=prey)

agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
agent_comp.add_node(player)
agent_comp.add_node(prey)
agent_comp.add_node(greedy_action_mech)

agent_comp.env = ForagerEnv() # NEW: ForagerEnv must be stored in an attribute on the Composition

def main():

    # NEW: get_next_input interactively returns a new input from the ForagerEnv
    #      (rather than specifying a pre-determined list of input values)
    def get_next_input(env, result):
        action = np.where(result[0] == 0, 0, result[0] / np.abs(result[0]))
        env_step = env.step(action)
        observation = env_step[0]
        done = env_step[2]
        if not done:
            # NEW: This function MUST return a dictionary of input values for a single trial for each INPUT node
            return {player: [observation[player_coord_idx]],
                    prey: [observation[prey_coord_idx]]}
        return done

    if RENDER:
        agent_comp.env.render()

    BIN_EXECUTE = 'Python'
    if PNL_COMPILE:
        BIN_EXECUTE = 'LLVM'

    max_steps = 100 # maximum number of steps before agent quits
    num_games = 3
    for i in range(3):
        agent_comp.run(inputs=get_next_input,     # specify the NAME of a fn in lieu of an inputs dict
                       num_trials=max_steps,      # maximum number of steps before agent quits
                       bin_execute=BIN_EXECUTE)

if __name__ == "__main__":
    main()



import argparse

import numpy as np
import psyneulink as pnl

parser = argparse.ArgumentParser()
parser.add_argument('--no-plot', action='store_false', help='Disable plotting', dest='enable_plot')
parser.add_argument('--threshold', type=float, help='Termination threshold for response output (default: %(default)f)', default=0.70)
parser.add_argument('--settle-trials', type=int, help='Number of trials for composition to initialize and settle (default: %(default)d)', default=200)
args = parser.parse_args()

# Implements the Kalanthroff, Davelaar, Henik, Goldfarb & Usher model: Task Conflict and Proactive Control:
# A Computational Theory of the Stroop Task. Psychol Rev. 2018 Jan;125(1):59-82. doi: 10.1037/rev0000083.
# Epub 2017 Oct 16.
# #https://www.ncbi.nlm.nih.gov/pubmed/29035077

# Define Variables ----------------------------------------------------------------------------------------------------
Lambda = 0.03            # PsyNeuLink has Euler integration constant reversed (1-0.97)
pc_high = 0.15           # High proactive control from Figure 6 in Paper
pc_low = 0.025           # Low proactive control from Figure 6 in Paper
pc = pc_low              # Select proactive control
inhibition = -1.3        # Inhibition between units within a layer
inhibition_task = -1.9   # Inhibition between units within task layer
bias = -0.3              # bias input to color feature layer and word feature layer
threshold = args.threshold
settle = args.settle_trials    # Number of trials until Composition settles

# Create mechanisms ---------------------------------------------------------------------------------------------------
# 4 Input layers for color, word, task & bias
colors_input_layer = pnl.TransferMechanism(
    size=2,
    function=pnl.Linear,
    name='COLORS_INPUT'
)

words_input_layer = pnl.TransferMechanism(
    size=2,
    function=pnl.Linear,
    name='WORDS_INPUT'
)

task_input_layer = pnl.TransferMechanism(
    size=2,
    function=pnl.Linear,
    name='PROACTIVE_CONTROL'
)

bias_input = pnl.TransferMechanism(
    size=2,
    function=pnl.Linear,
    name='BIAS'
)

# Built python function to ensure that the logistic function outputs 0 when input is <= 0


def my_special_Logistic(variable):
    maxi = variable - 0.0180
    output = np.fmax([0], maxi)
    return output

# Built python function that takes output of special logistic function and computes conflict by multiplying
# output both task units with each over times 500


def my_conflict_function(variable):
    maxi = variable - 0.0180
    new = np.fmax([0], maxi)
    out = [new[0] * new[1] * 500]
    return out


# Create color feature layer, word feature layer, task demand layer and response layer
color_feature_layer = pnl.RecurrentTransferMechanism(
    size=2,                     # Define unit size
    function=pnl.Logistic(gain=4, x_0=1),       # to 4 & bias to 1
    integrator_mode=True,       # Set IntegratorFunction mode to True
    integration_rate=Lambda,    # smoothing factor ==  integration rate
    hetero=inhibition,          # Inhibition among units within a layer
    output_ports=[{                          # Create new OutputPort by applying
        pnl.NAME: 'SPECIAL_LOGISTIC',         # the "my_special_Logistic" function
        pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
        pnl.FUNCTION: my_special_Logistic
    }],
    name='COLOR_LAYER')

# The word_feature_layer is set up as the color_feature_layer
word_feature_layer = pnl.RecurrentTransferMechanism(
    size=2,                     # Define unit size
    function=pnl.Logistic(gain=4, x_0=1),            # to 4 & bias to 1
    integrator_mode=True,   # Set IntegratorFunction mode to True
    integration_rate=Lambda,  # smoothing factor ==  integration rate
    hetero=inhibition,      # Inhibition among units within a layer
    output_ports=[{              # Create new OutputPort by applying
        pnl.NAME: 'SPECIAL_LOGISTIC',        # the "my_special_Logistic" function
        pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
        pnl.FUNCTION: my_special_Logistic
    }],
    name='WORD_LAYER')


# The response_layer is set up as the color_feature_layer & the word_feature_layer
response_layer = pnl.RecurrentTransferMechanism(
    size=2,                         # Define unit size
    function=pnl.Logistic(gain=4, x_0=1),           # to 4 & bias to 1
    integrator_mode=True,           # Set IntegratorFunction mode to True
    integration_rate=Lambda,        # smoothing factor ==  integration rate
    hetero=inhibition,              # Inhibition among units within a layer
    output_ports=[{           # Create new OutputPort by applying
        pnl.NAME: 'SPECIAL_LOGISTIC',        # the "my_special_Logistic" function
        pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
        pnl.FUNCTION: my_special_Logistic
    }],
    name='RESPONSE_LAYER'
)

# The task_demand_layer is set up as the color_feature_layer but with a different python function on it's OutputPort
# and a differnet inhibition weight on the hetero
task_demand_layer = pnl.RecurrentTransferMechanism(
    size=2,                      # Define unit size
    function=pnl.Logistic(gain=4, x_0=1),            # to 4 & bias to 1
    integrator_mode=True,   # Set IntegratorFunction mode to True
    integration_rate=Lambda,  # smoothing factor ==  integration rate
    hetero=inhibition_task,  # Inhibition among units within a layer
    output_ports=[               # Create new OutputPort by applying
        {
            pnl.NAME: 'SPECIAL_LOGISTIC',        # the "my_conflict_function" function
            pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
            pnl.FUNCTION: my_special_Logistic
        },
        {
            pnl.NAME: 'CONFLICT',
            pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
            pnl.FUNCTION: my_conflict_function
        }
    ],
    name='TASK_LAYER'
)


# Log mechanisms ------------------------------------------------------------------------------------------------------
color_feature_layer.set_log_conditions('SPECIAL_LOGISTIC')       # Log output of my_special_Logistic function
word_feature_layer.set_log_conditions('SPECIAL_LOGISTIC')        # Log output of my_special_Logistic function
response_layer.set_log_conditions('SPECIAL_LOGISTIC')            # Log output of my_special_Logistic function
task_demand_layer.set_log_conditions('SPECIAL_LOGISTIC')         # Log output of my_special_Logistic function

task_demand_layer.set_log_conditions('CONFLICT')                 # Log outout of my_conflict_function function

# Connect mechanisms --------------------------------------------------------------------------------------------------
color_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0],     # response layer projections are set to all
        [0.0, 0.0]      # zero for initialization period first
    ])
)
word_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ])
)
color_task_weights = pnl.MappingProjection(
    matrix=np.array([
        [2.0, 0.0],     # color to task projection
        [2.0, 0.0]
    ])
)
word_task_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 2.0],     # word to task projection
        [0.0, 2.0]
    ])
)
task_color_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.0, 1.0],     # task to color projection
        [0.0, 0.0]
    ])
)
task_word_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0],     # task to word projection
        [1.0, 1.0]
    ])
)
color_response_weights = pnl.MappingProjection(
    matrix=np.array([
        [2.0, 0.0],     # color to response projection
        [0.0, 2.0]
    ])
)
word_response_weights = pnl.MappingProjection(
    matrix=np.array([
        [2.5, 0.0],     # word to response projection
        [0.0, 2.5]
    ])
)
task_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.0, 0.0],     # proactive control to task
        [0.0, 1.0]
    ])
)

# to send a control signal from the task demand layer to the response layer,
# set matrix to -1 to reduce response layer activation
# specify the sender of the projection which is the second OutputPort the task demand layer
# specify the receiver of the projection
task_conflict_to_response_weights = pnl.MappingProjection(
    matrix=np.array([[-1.0, -1.0]]),
    sender=task_demand_layer.output_ports[1],
    receiver=response_layer
)

# Create pathways -----------------------------------------------------------------------------------------------------
color_response_process = pnl.Pathway(pathway=[
    colors_input_layer,
    color_input_weights,
    color_feature_layer,
    color_response_weights,
    response_layer
],
    name='COLORS_RESPONSE_PROCESS'
)

word_response_process = pnl.Pathway(
    pathway=[words_input_layer,
             word_input_weights,
             word_feature_layer,
             word_response_weights,
             response_layer
             ],
    name='WORDS_RESPONSE_PROCESS'
)

task_color_process = pnl.Pathway(
    pathway=[task_input_layer,
             task_input_weights,
             task_demand_layer,
             task_color_weights,
             color_feature_layer,
             color_task_weights,
             task_demand_layer
             ],
    name='TASK_COLOR_PROCESS'
)

task_word_process = pnl.Pathway(
    pathway=[task_input_layer,
             task_demand_layer,
             task_word_weights,
             word_feature_layer,
             word_task_weights,
             task_demand_layer
             ],
    name='TASK_WORD_PROCESS'
)

bias_color_process = pnl.Pathway(
    pathway=[bias_input, color_feature_layer],
    name='BIAS_COLOR'
)

bias_word_process = pnl.Pathway(
    pathway=[bias_input, word_feature_layer],
    name='WORD_COLOR'
)

conflict_process = pnl.Pathway(
    pathway=[
        task_demand_layer,
        task_conflict_to_response_weights,
        response_layer
    ],
    name='CONFLICT_PROCESS'
)

# Create Composition --------------------------------------------------------------------------------------------------
PCTC = pnl.Composition(
    pathways=[
        word_response_process,
        color_response_process,
        task_color_process,
        task_word_process,
        bias_word_process,
        bias_color_process,
        task_word_process,
        conflict_process
    ],
    reinitialize_mechanisms_when=pnl.Never(),
    name='PCTC_MODEL')

# Create threshold function -------------------------------------------------------------------------------------------

def pass_threshold(response_layer, thresh, context):
    results1 = response_layer.get_output_values(context)[0][0]  # red response
    results2 = response_layer.get_output_values(context)[0][1]  # green response
    # print(results1)
    # print(results2)
    if results1 >= thresh or results2 >= thresh:
        return True
    return False


terminate_trial = {
    pnl.TimeScale.TRIAL: pnl.While(pass_threshold, response_layer, threshold)
}

# Create test trials function -----------------------------------------------------------------------------------------
# a BLUE word input is [1,0] to words_input_layer and GREEN word is [0,1]
# a blue color input is [1,0] to colors_input_layer and green color is [0,1]
# a color-naming trial is [1,0] to task_layer and a word-reading trial is [0,1]


def trial_dict(blue_color, green_color, blue_word, green_word, PC_CN, PC_WR, bias):
    trialdict = {
        colors_input_layer: [blue_color, green_color],
        words_input_layer: [blue_word, green_word],
        task_input_layer: [PC_CN, PC_WR],
        bias_input: [bias, bias]
    }
    return trialdict


initialize_input = trial_dict(1.0, 0.0, 1.0, 0.0, pc, 0.0, bias)

# Run congruent trial -------------------------------------------------------------------------------------------------
congruent_input = trial_dict(1.0, 0.0, 1.0, 0.0, pc, 0.0, bias)  # specify congruent trial input
# run Composition to settle for 200 trials with congruent stimuli input
PCTC.run(inputs=initialize_input, num_trials=settle)

color_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],      # set color input projections to 1 on the diagonals to e.g.
        [0.0, 1.0]
    ]),     # send a green color input to the green unit of the color layer
    PCTC
)
word_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],      # the same for word input projections
        [0.0, 1.0]
    ]),
    PCTC
)

# run Composition with congruent stimulus input until threshold in of of the response layer units is reached
PCTC.run(inputs=congruent_input, termination_processing=terminate_trial)

# Store values from run -----------------------------------------------------------------------------------------------
t = task_demand_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')    # Log task output from special logistic function
tt = t[PCTC.name]['SPECIAL_LOGISTIC']
n_con = tt.shape[0]
ttt_cong = tt.reshape(n_con, 2)
conflict_con = ttt_cong[200:, 0] * ttt_cong[200:, 1] * 100             # Compute conflict for plotting (as in MATLAB code)

c = color_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')  # Log color output from special logistic function
cc = c[PCTC.name]['SPECIAL_LOGISTIC']
ccc_cong = cc.reshape(n_con, 2)
w = word_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')   # Log word output from special logistic function
ww = w[PCTC.name]['SPECIAL_LOGISTIC']
www_cong = ww.reshape(n_con, 2)
r = response_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')       # Log response output from special logistic function
rr = r[PCTC.name]['SPECIAL_LOGISTIC']
rrr_cong = rr.reshape(n_con, 2)

# Clear log & reset --------------------------------------------------------------------------------------------
response_layer.log.clear_entries()
color_feature_layer.log.clear_entries()
word_feature_layer.log.clear_entries()
task_demand_layer.log.clear_entries()

color_feature_layer.reset([[0, 0]], context=PCTC)
word_feature_layer.reset([[0, 0]], context=PCTC)
response_layer.reset([[0, 0]], context=PCTC)
task_demand_layer.reset([[0, 0]], context=PCTC)

# Run neutral trials --------------------------------------------------------------------------------------------------
# Set input projections back to 0 for settling period
color_input_weights.parameters.matrix.set(
    np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ]),
    PCTC
)
word_input_weights.parameters.matrix.set(
    np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ]),
    PCTC
)

neutral_input = trial_dict(1.0, 0.0, 0.0, 0.0, pc, 0.0, bias)  # create neutral stimuli input
# run Compositoin to settle for 200 trials with neutral stimuli input
PCTC.run(inputs=initialize_input, num_trials=settle)

color_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],      # Set input projections to 1 for stimulus presentation period
        [0.0, 1.0]
    ]),
    PCTC
)
word_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ]),
    PCTC
)

# run Composition with neutral stimulus input until threshold in of of the response layer units is reached
PCTC.run(inputs=neutral_input, termination_processing=terminate_trial)

# Store values from neutral run ---------------------------------------------------------------------------------------
t = task_demand_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
tt = t[PCTC.name]['SPECIAL_LOGISTIC']
n_neutral = tt.shape[0]
ttt_neutral = tt.reshape(n_neutral, 2)
conflict_neutral = ttt_neutral[200:, 0] * ttt_neutral[200:, 1] * 100

c = color_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
cc = c[PCTC.name]['SPECIAL_LOGISTIC']
ccc_neutral = cc.reshape(n_neutral, 2)
w = word_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
ww = w[PCTC.name]['SPECIAL_LOGISTIC']
www_neutral = ww.reshape(n_neutral, 2)
r = response_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
rr = r[PCTC.name]['SPECIAL_LOGISTIC']
rrr_neutral = rr.reshape(n_neutral, 2)
# Clear log & reset --------------------------------------------------------------------------------------------

response_layer.log.clear_entries()
color_feature_layer.log.clear_entries()
word_feature_layer.log.clear_entries()
task_demand_layer.log.clear_entries()

color_feature_layer.reset([[0, 0]], context=PCTC)
word_feature_layer.reset([[0, 0]], context=PCTC)
response_layer.reset([[0, 0]], context=PCTC)
task_demand_layer.reset([[0, 0]], context=PCTC)

# Run incongruent trials ----------------------------------------------------------------------------------------------
# Set input projections back to 0 for settling period
color_input_weights.parameters.matrix.set(
    np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ]),
    PCTC
)
word_input_weights.parameters.matrix.set(
    np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ]),
    PCTC
)

incongruent_input = trial_dict(1.0, 0.0, 0.0, 1.0, pc, 0.0, bias)
PCTC.run(inputs=initialize_input, num_trials=settle)

color_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ]),
    PCTC
)
word_input_weights.parameters.matrix.set(
    np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ]),
    PCTC
)

PCTC.run(inputs=incongruent_input, termination_processing=terminate_trial)

# Store values from neutral run ---------------------------------------------------------------------------------------

t = task_demand_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
tt = t[PCTC.name]['SPECIAL_LOGISTIC']
n_incon = tt.shape[0]
ttt_incong = tt.reshape(n_incon, 2)
conflict_incon = ttt_incong[200:, 0] * ttt_incong[200:, 1] * 100

c = color_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
cc = c[PCTC.name]['SPECIAL_LOGISTIC']
ccc_incong = cc.reshape(n_incon, 2)
w = word_feature_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
ww = w[PCTC.name]['SPECIAL_LOGISTIC']
www_incong = ww.reshape(n_incon, 2)
r = response_layer.log.nparray_dictionary('SPECIAL_LOGISTIC')
rr = r[PCTC.name]['SPECIAL_LOGISTIC']
rrr_incong = rr.reshape(n_incon, 2)

# Plotting ------------------------------------------------------------------------------------------------------------
if args.enable_plot:
    import matplotlib.pyplot as plt

    # Set up plot structure
    fig, axes = plt.subplots(nrows=3, ncols=4, sharey=True, sharex=True)
    axes[0, 0].set_ylabel('Congruent')
    axes[1, 0].set_ylabel('Neutral')
    axes[2, 0].set_ylabel('Incongruent')

    axes[0, 0].set_title('Task demand units', fontsize=9)
    axes[0, 1].set_title('Response units', fontsize=9)
    axes[0, 2].set_title('Color feature map', fontsize=9)
    axes[0, 3].set_title('Word feature map', fontsize=9)
    plt.setp(
        axes,
        xticks=[0, 400, 780],
        yticks=[0, 0.4, 0.79],
        yticklabels=['0', '0.4', '0.8'],
        xticklabels=['0', '400', '800']
    )

    # Plot congruent output --------------------------
    axes[0, 0].plot(ttt_cong[settle:, 0], 'c')
    axes[0, 0].plot(ttt_cong[settle:, 1], 'k')
    axes[0, 0].plot(conflict_con, 'r')

    axes[0, 1].plot(rrr_cong[settle:, 0], 'b')
    axes[0, 1].plot(rrr_cong[settle:, 1], 'g')
    axes[0, 1].plot([0, n_con - settle], [threshold, threshold], 'k')
    axes[0, 2].plot(ccc_cong[settle:, 0], 'b')
    axes[0, 2].plot(ccc_cong[settle:, 1], 'g')

    axes[0, 3].plot(www_cong[settle:, 0], 'b')
    axes[0, 3].plot(www_cong[settle:, 1], 'g')

    # Plot neutral output --------------------------
    axes[1, 0].plot(ttt_neutral[settle:, 0], 'c')
    axes[1, 0].plot(ttt_neutral[settle:, 1], 'k')
    axes[1, 0].plot(conflict_neutral, 'r')

    axes[1, 1].plot(rrr_neutral[settle:, 0], 'b')
    axes[1, 1].plot(rrr_neutral[settle:, 1], 'g')
    axes[1, 1].plot([0, n_neutral - settle], [threshold, threshold], 'k')
    axes[1, 2].plot(ccc_neutral[settle:, 0], 'b')
    axes[1, 2].plot(ccc_neutral[settle:, 1], 'g')

    axes[1, 3].plot(www_neutral[settle:, 0], 'b')
    axes[1, 3].plot(www_neutral[settle:, 1], 'g')

    # Plot incongruent output --------------------------
    axes[2, 0].plot(ttt_incong[settle:, 0], 'c')
    axes[2, 0].plot(ttt_incong[settle:, 1], 'k')
    axes[2, 0].plot(conflict_incon, 'r')

    axes[2, 1].plot(rrr_incong[settle:, 0], 'b')
    axes[2, 1].plot(rrr_incong[settle:, 1], 'g')
    axes[2, 1].plot([0, n_incon - settle], [threshold, threshold], 'k')
    axes[2, 2].plot(ccc_incong[settle:, 0], 'b')
    axes[2, 2].plot(ccc_incong[settle:, 1], 'g')

    axes[2, 3].plot(www_incong[settle:, 0], 'b')
    axes[2, 3].plot(www_incong[settle:, 1], 'g')

    plt.show(block=not pnl._called_from_pytest)



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *****************************************  STATEFUL FUNCTION *********************************************************
"""

* `StatefulFunction`
* `IntegratorFunctions`
* `MemoryFunctions`

"""

import abc
import collections
import typecheck as tc
import warnings
import numbers

import numpy as np

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.component import DefaultsFlexibility, _has_initializers_setter, ComponentsMeta
from psyneulink.core.components.functions.function import Function_Base, FunctionError
from psyneulink.core.components.functions.distributionfunctions import DistributionFunction
from psyneulink.core.globals.keywords import STATEFUL_FUNCTION_TYPE, STATEFUL_FUNCTION, NOISE, RATE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.utilities import parameter_spec, iscompatible, object_has_single_value, convert_to_np_array, contains_type
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.context import ContextFlags, handle_external_context

__all__ = ['StatefulFunction']


class StatefulFunction(Function_Base): #  ---------------------------------------------------------------------
    """
    StatefulFunction(           \
        default_variable=None,  \
        initializer,            \
        rate=1.0,               \
        noise=0.0,              \
        params=None,            \
        owner=None,             \
        prefs=None,             \
        )

    .. _StatefulFunction:

    Abstract base class for Functions the result of which depend on their `previous_value
    <StatefulFunction.previous_value>` attribute.

    COMMENT:
    NARRATIVE HERE THAT EXPLAINS:
    A) initializers and stateful_attributes
    B) initializer (note singular) is a prespecified member of initializers
       that contains the value with which to initiailzer previous_value
    COMMENT


    Arguments
    ---------

    default_variable : number, list or array : default class_defaults.variable
        specifies a template for `variable <StatefulFunction.variable>`.

    initializer : float, list or 1d array : default 0.0
        specifies initial value for `previous_value <StatefulFunction.previous_value>`.  If it is a list or array,
        it must be the same length as `variable <StatefulFunction.variable>` (see `initializer
        <StatefulFunction.initializer>` for details).

    rate : float, list or 1d array : default 1.0
        specifies value used as a scaling parameter in a subclass-dependent way (see `rate <StatefulFunction.rate>` for
        details); if it is a list or array, it must be the same length as `variable <StatefulFunction.default_variable>`.

    noise : float, function, list or 1d array : default 0.0
        specifies random value added in each call to `function <StatefulFunction.function>`; if it is a list or
        array, it must be the same length as `variable <StatefulFunction.default_variable>` (see `noise
        <StatefulFunction.noise>` for details).

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or array
        current input value.

    initializer : float or 1d array
        determines initial value assigned to `previous_value <StatefulFunction.previous_value>`. If `variable
        <StatefulFunction.variable>` is a list or array, and initializer is a float or has a single element, it is
        applied to each element of `previous_value <StatefulFunction.previous_value>`. If initializer is a list or
        array,each element is applied to the corresponding element of `previous_value <Integrator.previous_value>`.

    previous_value : 1d array
        last value returned (i.e., for which state is being maintained).

    initializers : list
        stores the names of the initialization attributes for each of the stateful attributes of the function. The
        index i item in initializers provides the initialization value for the index i item in `stateful_attributes
        <StatefulFunction.stateful_attributes>`.

    stateful_attributes : list
        stores the names of each of the stateful attributes of the function. The index i item in stateful_attributes is
        initialized by the value of the initialization attribute whose name is stored in index i of `initializers
        <StatefulFunction.initializers>`. In most cases, the stateful_attributes, in that order, are the return values
        of the function.

    .. _Stateful_Rate:

    rate : float or 1d array
        on each call to `function <StatefulFunction.function>`, applied to `variable <StatefulFunction.variable>`,
        `previous_value <StatefulFunction.previous_value>`, neither, or both, depending on implementation by
        subclass.  If it is a float or has a single value, it is applied to all elements of its target(s);  if it has
        more than one element, each element is applied to the corresponding element of its target(s).

    .. _Stateful_Noise:

    noise : float, function, list, or 1d array
        random value added on each call to `function <StatefulFunction.function>`. If `variable
        <StatefulFunction.variable>` is a list or array, and noise is a float or function, it is applied
        for each element of `variable <StatefulFunction.variable>`. If noise is a function, it is executed and applied
        separately for each element of `variable <StatefulFunction.variable>`.  If noise is a list or array,
        it is applied elementwise (i.e., in Hadamard form).

        .. hint::
            To generate random noise that varies for every execution, a probability distribution function should be
            used (see `Distribution Functions <DistributionFunction>` for details), that generates a new noise value
            from its distribution on each execution. If noise is specified as a float, a function with a fixed
            output, or a list or array of either of these, then noise is simply an offset that remains the same
            across all executions.

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the Function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentType = STATEFUL_FUNCTION_TYPE
    componentName = STATEFUL_FUNCTION

    class Parameters(Function_Base.Parameters):
        """
            Attributes
            ----------

                initializer
                    see `initializer <StatefulFunction.initializer>`

                    :default value: numpy.array([0])
                    :type: ``numpy.ndarray``

                noise
                    see `noise <StatefulFunction.noise>`

                    :default value: 0.0
                    :type: ``float``

                previous_value
                    see `previous_value <StatefulFunction.previous_value>`

                    :default value: numpy.array([0])
                    :type: ``numpy.ndarray``

                rate
                    see `rate <StatefulFunction.rate>`

                    :default value: 1.0
                    :type: ``float``
        """
        noise = Parameter(0.0, modulable=True)
        rate = Parameter(1.0, modulable=True)
        previous_value = Parameter(np.array([0]), initializer='initializer', pnl_internal=True)
        initializer = Parameter(np.array([0]), pnl_internal=True)
        has_initializers = Parameter(True, setter=_has_initializers_setter, pnl_internal=True)

        def _validate_noise(self, noise):
            if (
                isinstance(noise, collections.Iterable)
                # assume ComponentsMeta are functions
                and contains_type(noise, ComponentsMeta)
            ):
                # TODO: make this validation unnecessary by handling automatically?
                return 'functions in a list must be instantiated and have the desired noise variable shape'

    @handle_external_context()
    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 rate=None,
                 noise=None,
                 initializer=None,
                 params: tc.optional(tc.optional(dict)) = None,
                 owner=None,
                 prefs: tc.optional(is_pref_set) = None,
                 context=None,
                 **kwargs
                 ):

        if not hasattr(self, "initializers"):
            self.initializers = ["initializer"]

        if not hasattr(self, "stateful_attributes"):
            self.stateful_attributes = ["previous_value"]

        super().__init__(
            default_variable=default_variable,
            rate=rate,
            initializer=initializer,
            noise=noise,
            params=params,
            owner=owner,
            prefs=prefs,
            context=context,
            **kwargs
        )

    def _validate(self, context=None):
        self._validate_rate(self.defaults.rate)
        self._validate_initializers(self.defaults.variable, context=context)
        super()._validate(context=context)

    def _validate_params(self, request_set, target_set=None, context=None):

        # Handle list or array for rate specification
        if RATE in request_set:
            rate = request_set[RATE]

            if isinstance(rate, (list, np.ndarray)) and not iscompatible(rate, self.defaults.variable):
                if len(rate) != 1 and len(rate) != np.array(self.defaults.variable).size:
                    # If the variable was not specified, then reformat it to match rate specification
                    #    and assign class_defaults.variable accordingly
                    # Note: this situation can arise when the rate is parametrized (e.g., as an array) in the
                    #       StatefulFunction's constructor, where that is used as a specification for a function parameter
                    #       (e.g., for an IntegratorMechanism), whereas the input is specified as part of the
                    #       object to which the function parameter belongs (e.g., the IntegratorMechanism); in that
                    #       case, the StatefulFunction gets instantiated using its class_defaults.variable ([[0]]) before
                    #       the object itself, thus does not see the array specification for the input.
                    if self._variable_shape_flexibility is DefaultsFlexibility.FLEXIBLE:
                        self._instantiate_defaults(variable=np.zeros_like(np.array(rate)), context=context)
                        if self.verbosePref:
                            warnings.warn(
                                "The length ({}) of the array specified for the rate parameter ({}) of {} "
                                "must match the length ({}) of the default input ({});  "
                                "the default input has been updated to match".format(
                                    len(rate),
                                    rate,
                                    self.name,
                                    np.array(self.defaults.variable).size
                                ),
                                self.defaults.variable,
                            )
                    else:
                        raise FunctionError(
                            "The length of the array specified for the rate parameter of {} ({}) "
                            "must match the length of the default input ({}).".format(
                                self.name,
                                # rate,
                                len(rate),
                                np.array(self.defaults.variable).size,
                                # self.defaults.variable,
                            )
                        )

        super()._validate_params(request_set=request_set,
                                 target_set=target_set,
                                 context=context)

        if NOISE in target_set:
            noise = target_set[NOISE]
            if isinstance(noise, DistributionFunction):
                noise.owner = self
                target_set[NOISE] = noise.execute
            self._validate_noise(target_set[NOISE])

    def _validate_initializers(self, default_variable, context=None):
        for initial_value_name in self.initializers:

            initial_value = self._get_current_parameter_value(initial_value_name, context=context)

            if isinstance(initial_value, (list, np.ndarray)):
                if len(initial_value) != 1:
                    # np.atleast_2d may not be necessary here?
                    if np.shape(np.atleast_2d(initial_value)) != np.shape(np.atleast_2d(default_variable)):
                        raise FunctionError("{}'s {} ({}) is incompatible with its default_variable ({}) ."
                                            .format(self.name, initial_value_name, initial_value, default_variable))
            elif not isinstance(initial_value, (float, int)):
                raise FunctionError("{}'s {} ({}) must be a number or a list/array of numbers."
                                    .format(self.name, initial_value_name, initial_value))

    def _validate_rate(self, rate):
        # FIX: CAN WE JUST GET RID OF THIS?
        # kmantel: this duplicates much code in _validate_params above, but that calls _instantiate_defaults
        # which I don't think is the right thing to do here, but if you don't call it in _validate_params
        # then a lot of things don't get instantiated properly
        if rate is not None:
            if isinstance(rate, list):
                rate = np.asarray(rate)

            rate_type_msg = 'The rate parameter of {0} must be a number or an array/list of at most 1d (you gave: {1})'
            if isinstance(rate, np.ndarray):
                # kmantel: current test_gating test depends on 2d rate
                #   this should be looked at but for now this restriction is removed
                # if rate.ndim > 1:
                #     raise FunctionError(rate_type_msg.format(self.name, rate))
                pass
            elif not isinstance(rate, numbers.Number):
                raise FunctionError(rate_type_msg.format(self.name, rate))

            if isinstance(rate, np.ndarray) and not iscompatible(rate, self.defaults.variable):
                if len(rate) != 1 and len(rate) != np.array(self.defaults.variable).size:
                    if self._variable_shape_flexibility is DefaultsFlexibility.FLEXIBLE:
                        self.defaults.variable = np.zeros_like(np.array(rate))
                        if self.verbosePref:
                            warnings.warn(
                                "The length ({}) of the array specified for the rate parameter ({}) of {} "
                                "must match the length ({}) of the default input ({});  "
                                "the default input has been updated to match".format(
                                    len(rate),
                                    rate,
                                    self.name,
                                    np.array(self.defaults.variable).size
                                ),
                                self.defaults.variable,
                            )
                        self._instantiate_value()
                        self._variable_shape_flexibility = DefaultsFlexibility.INCREASE_DIMENSION
                    else:
                        raise FunctionError(
                            "The length of the array specified for the rate parameter of {} ({})"
                            "must match the length of the default input ({}).".format(
                                len(rate),
                                # rate,
                                self.name,
                                np.array(self.defaults.variable).size,
                                # self.defaults.variable,
                            )
                        )

    # Ensure that the noise parameter makes sense with the input type and shape; flag any noise functions that will
    # need to be executed
    def _validate_noise(self, noise):
        # Noise is a list or array
        if isinstance(noise, (np.ndarray, list)):
            if len(noise) == 1:
                pass
            # Variable is a list/array
            elif (not iscompatible(np.atleast_2d(noise), self.defaults.variable)
                  and not iscompatible(np.atleast_1d(noise), self.defaults.variable) and len(noise) > 1):
                raise FunctionError(
                    "Noise parameter ({}) does not match default variable ({}). Noise parameter of {} "
                    "must be specified as a float, a function, or an array of the appropriate shape ({}).".format(
                        noise, self.defaults.variable, self.name,
                        np.shape(np.array(self.defaults.variable))
                    ),
                    component=self
                )
            else:
                for i in range(len(noise)):
                    if isinstance(noise[i], DistributionFunction):
                        noise[i] = noise[i].execute
                    # if not isinstance(noise[i], (float, int)) and not callable(noise[i]):
                    if not np.isscalar(noise[i]) and not callable(noise[i]):
                        raise FunctionError("The elements of a noise list or array must be scalars or functions. "
                                            "{} is not a valid noise element for {}".format(noise[i], self.name))

    def _try_execute_param(self, param, var, context=None):

        # FIX: [JDC 12/18/18 - HACK TO DEAL WITH ENFORCEMENT OF 2D BELOW]
        param_shape = np.array(param).shape
        if not len(param_shape):
            param_shape = np.array(var).shape
        # param is a list; if any element is callable, execute it
        if isinstance(param, (np.ndarray, list)):
            # NOTE: np.atleast_2d will cause problems if the param has "rows" of different lengths
            # FIX: WHY FORCE 2d??
            param = np.atleast_2d(param)
            for i in range(len(param)):
                for j in range(len(param[i])):
                    try:
                        param[i][j] = param[i][j](context=context)
                    except TypeError:
                        try:
                            param[i][j] = param[i][j]()
                        except TypeError:
                            pass
            try:
                param = param.reshape(param_shape)
            except ValueError:
                if object_has_single_value(param):
                    param = np.full(param_shape, float(param))

        # param is one function
        elif callable(param):
            # NOTE: np.atleast_2d will cause problems if the param has "rows" of different lengths
            new_param = []
            # FIX: WHY FORCE 2d??
            for row in np.atleast_2d(var):
            # for row in np.atleast_1d(var):
            # for row in var:
                new_row = []
                for item in row:
                    try:
                        val = param(context=context)
                    except TypeError:
                        val = param()
                    new_row.append(val)
                new_param.append(new_row)
            param = np.asarray(new_param)
            # FIX: [JDC 12/18/18 - HACK TO DEAL WITH ENFORCEMENT OF 2D ABOVE]
            try:
                if len(np.squeeze(param)):
                    param = param.reshape(param_shape)
            except TypeError:
                pass

        return param

    def _instantiate_attributes_before_function(self, function=None, context=None):
        if not self.parameters.initializer._user_specified:
            self._initialize_previous_value(np.zeros_like(self.defaults.variable), context)

        # use np.broadcast_to to guarantee that all initializer type attributes take on the same shape as variable
        if not np.isscalar(self.defaults.variable):
            for attr in self.initializers:
                param = getattr(self.parameters, attr)
                param._set(
                    np.broadcast_to(
                        param._get(context),
                        self.defaults.variable.shape
                    ).copy(),
                    context
                )

        # create all stateful attributes and initialize their values to the current values of their
        # corresponding initializer attributes
        for attr_name in self.stateful_attributes:
            initializer_value = getattr(self.parameters, getattr(self.parameters, attr_name).initializer)._get(context).copy()
            getattr(self.parameters, attr_name)._set(initializer_value, context)

        super()._instantiate_attributes_before_function(function=function, context=context)

    def _initialize_previous_value(self, initializer, context=None):
        initializer = convert_to_np_array(initializer, dimension=1)

        self.defaults.initializer = initializer.copy()
        self.parameters.initializer._set(initializer.copy(), context)

        self.defaults.previous_value = initializer.copy()
        self.parameters.previous_value.set(initializer.copy(), context)

        return initializer

    @handle_external_context()
    def _update_default_variable(self, new_default_variable, context=None):
        if not self.parameters.initializer._user_specified:
            self._initialize_previous_value(np.zeros_like(new_default_variable), context)

        super()._update_default_variable(new_default_variable, context=context)

    def _parse_value_order(self, **kwargs):
        """
            Returns:
                tuple: the values of the keyword arguments in the order
                in which they appear in this Component's `value
                <Component.value>`
        """
        return tuple(v for k, v in kwargs.items())

    @handle_external_context(fallback_most_recent=True)
    def reset(self, *args, context=None, **kwargs):
        """
            Resets `value <StatefulFunction.previous_value>`  and `previous_value <StatefulFunction.previous_value>`
            to the specified value(s).

            If arguments are passed into the reset method, then reset sets each of the attributes in
            `stateful_attributes <StatefulFunction.stateful_attributes>` to the value of the corresponding argument.
            Next, it sets the `value <StatefulFunction.value>` to a list containing each of the argument values.

            If reset is called without arguments, then it sets each of the attributes in `stateful_attributes
            <StatefulFunction.stateful_attributes>` to the value of the corresponding attribute in `initializers
            <StatefulFunction.initializers>`. Next, it sets the `value <StatefulFunction.value>` to a list containing
            the values of each of the attributes in `initializers <StatefulFunction.initializers>`.

            Often, the only attribute in `stateful_attributes <StatefulFunction.stateful_attributes>` is
            `previous_value <StatefulFunction.previous_value>` and the only attribute in `initializers
            <StatefulFunction.initializers>` is `initializer <StatefulFunction.initializer>`, in which case
            the reset method sets `previous_value <StatefulFunction.previous_value>` and `value
            <StatefulFunction.value>` to either the value of the argument (if an argument was passed into
            reset) or the current value of `initializer <StatefulFunction.initializer>`.

            For specific types of StatefulFunction functions, the reset method may carry out other
            reinitialization steps.

        """
        num_stateful_attrs = len(self.stateful_attributes)
        if num_stateful_attrs >= 2:
            # old args specification can be supported only in subclasses
            # that explicitly define an order by overriding reset
            if len(args) > 0:
                raise FunctionError(
                    f'{self}.reset has more than one stateful attribute'
                    f' ({self.stateful_attributes}). You must specify reset'
                    ' values by keyword.'
                )
            if len(kwargs) != num_stateful_attrs:
                type_name = type(self).__name__
                raise FunctionError(
                    'StatefulFunction.reset must receive a keyword argument for'
                    f' each item in {type_name}.stateful_attributes in the order in'
                    f' which they appear in {type_name}.value'
                )

        if num_stateful_attrs == 1:
            try:
                kwargs[self.stateful_attributes[0]]
            except KeyError:
                try:
                    kwargs[self.stateful_attributes[0]] = args[0]
                except IndexError:
                    kwargs[self.stateful_attributes[0]] = None

        invalid_args = []

        # iterates in order arguments are sent in function call, so it
        # will match their order in value as long as they are listed
        # properly in subclass reset method signatures
        for attr in kwargs:
            try:
                kwargs[attr]
            except KeyError:
                kwargs[attr] = None

            if kwargs[attr] is not None:
                # from before: unsure if conversion to 1d necessary
                kwargs[attr] = np.atleast_1d(kwargs[attr])
            else:
                try:
                    kwargs[attr] = self._get_current_parameter_value(getattr(self.parameters, attr).initializer, context=context)
                except AttributeError:
                    invalid_args.append(attr)

        if len(invalid_args) > 0:
            raise FunctionError(
                f'Arguments {invalid_args} to reset are invalid because they do'
                f" not correspond to any of {self}'s stateful_attributes"
            )

        # rebuilding value rather than simply returning reinitialization_values in case any of the stateful
        # attrs are modified during assignment
        value = []
        for attr, v in kwargs.items():
            # FIXME: HACK: Do not reinitialize random_state
            if attr != "random_state":
                getattr(self.parameters, attr).set(kwargs[attr],
                                                   context, override=True)
                value.append(getattr(self.parameters, attr)._get(context))

        self.parameters.value.set(value, context, override=True)
        return value

    def _gen_llvm_function_reset(self, ctx, builder, params, state, arg_in, arg_out, *, tags:frozenset):
        assert "reset" in tags
        for a in self.stateful_attributes:
            initializer = getattr(self.parameters, a).initializer
            source_ptr = pnlvm.helpers.get_param_ptr(builder, self, params, initializer)
            dest_ptr = pnlvm.helpers.get_state_ptr(builder, self, state, a)
            if source_ptr.type != dest_ptr.type:
                warnings.warn("Shape mismatch: stateful param does not match the initializer: {}({}) vs. {}({})".format(initializer, source_ptr.type, a, dest_ptr.type))
                # Take a guess that dest just has an extra dimension
                assert len(dest_ptr.type.pointee) == 1
                dest_ptr = builder.gep(dest_ptr, [ctx.int32_ty(0),
                                                  ctx.int32_ty(0)])
            builder.store(builder.load(source_ptr), dest_ptr)

        return builder

    @abc.abstractmethod
    def _function(self, *args, **kwargs):
        raise FunctionError("StatefulFunction is not meant to be called explicitly")



import psyneulink as pnl
import numpy as np

# here we implement a test demo as in the EVC paper example:
#in v2 we add control signals and a EVC mechanism to the model

# EVC params for Umemoto et al

w_t = 0.065
w_d = 0.065#0.065 # made negative here to match -1 values for distractor
f_t = 1
f_d = 1


# EVC params for Umemoto et al
t0 = 0.2
c = 0.19
thresh = 0.21
x_0 = 0 # starting point

#wTarget = 0.065 # I think this has to do with learning and is constant over trials in Umemoto
costParam1 = 0.35
reconfCostParam1 = 5
#rewardTaskA = 50
#rewardTaskBToA = 0.7


# Control Parameters
signalSearchRange = pnl.SampleSpec(start=1.8, stop=2.2, step=0.2)
# signalSearchRange = pnl.SampleSpec(start=0.0, stop=0.4, step=0.2)

# Stimulus Mechanisms
Target_Stim = pnl.TransferMechanism(name='Target Stimulus', function=pnl.Linear)
Target_Stim.set_log_conditions('value')#, log_condition=pnl.PROCESSING) # Log Target_Rep

Distractor_Stim = pnl.TransferMechanism(name='Distractor Stimulus', function=pnl.Linear)
Distractor_Stim.set_log_conditions('value')#, log_condition=pnl.PROCESSING) # Log Target_Rep

# Processing Mechanisms (Control)
Target_Rep = pnl.TransferMechanism(name='Target Representation')

Target_Rep.set_log_conditions('value')#, log_condition=pnl.PROCESSING) # Log Target_Rep
Target_Rep.set_log_conditions('mod_slope')#, log_condition=pnl.PROCESSING)
Target_Rep.set_log_conditions('InputPort-0')#, log_condition=pnl.PROCESSING)

Distractor_Rep = pnl.TransferMechanism(name='Distractor Representation')

Distractor_Rep.set_log_conditions('value')#, log_condition=pnl.PROCESSING) # Log Flanker_Rep
Distractor_Rep.set_log_conditions('mod_slope')#, log_condition=pnl.PROCESSING)

# Processing Mechanism (Automatic)
Automatic_Component = pnl.TransferMechanism(name='Automatic Component')

Distactor_Component = pnl.TransferMechanism(name='Distactor Component')


Automatic_Component.set_log_conditions('value')#, log_condition=pnl.PROCESSING)

# Decision Mechanisms
Decision = pnl.DDM(function=pnl.DriftDiffusionAnalytical(
       # drift_rate=(0.1170),
        threshold=(thresh),
        noise=(c),
        starting_point=(x_0),
        t0=t0
    ),name='Decision',
    output_ports=[
        pnl.DECISION_VARIABLE,
        pnl.RESPONSE_TIME,
        pnl.PROBABILITY_UPPER_THRESHOLD,
        {
            pnl.NAME: 'OFFSET RT',
            pnl.VARIABLE: (pnl.OWNER_VALUE, 2),
            pnl.FUNCTION: pnl.Linear(0, slope=1.0, intercept=1)
        }
    ],) #drift_rate=(1.0),threshold=(0.2645),noise=(0.5),starting_point=(0), t0=0.15

Decision.set_log_conditions('InputPort-0')#, log_condition=pnl.PROCESSING)
Decision.set_log_conditions('PROBABILITY_UPPER_THRESHOLD')
print(Decision.loggable_items)
# Outcome Mechanisms:
Reward = pnl.TransferMechanism(name='Reward')

# Composition
Umemoto_comp = pnl.Composition(name="Umemoto_System")

#weights

Distractor_weight = pnl.MappingProjection(matrix=np.matrix([[-1]]),
                                     name='DISTRACTOR_WEIGHTS')
# ADD pathways
TargetControl_pathway = [Target_Stim, Target_Rep, Decision]
Umemoto_comp.add_linear_processing_pathway(TargetControl_pathway)

FlankerControl_pathway = [Distractor_Stim, Distractor_Rep, Distactor_Component, Distractor_weight, Decision]
Umemoto_comp.add_linear_processing_pathway(FlankerControl_pathway)

TargetAutomatic_pathway = [Target_Stim, Decision]
Umemoto_comp.add_linear_processing_pathway(TargetAutomatic_pathway)

FlankerAutomatic_pathway = [Distractor_Stim, Distactor_Component]
Umemoto_comp.add_linear_processing_pathway(FlankerAutomatic_pathway)

Reward_pathway = [Reward]
Umemoto_comp.add_linear_processing_pathway(Reward_pathway)

Umemoto_comp.add_node(Decision,
                      # required_roles=pnl.NodeRole.OUTPUT
                      )

# COMPOSITION
Target_Rep_Control_Signal = pnl.ControlSignal(modulates=[(pnl.SLOPE, Target_Rep)],
                                              function=pnl.Linear,
                                              variable=1.0,
                                              cost_options=[pnl.CostFunctions.INTENSITY, pnl.CostFunctions.ADJUSTMENT],
                                              intensity_cost_function=pnl.Exponential(scale=1, rate=1),
                                              # compute_reconfiguration_cost=pnl.Distance(metric=pnl.EUCLIDEAN),
                                              # adjustment_cost_function=pnl.Exponential(scale=1, rate=1, offset=-1),#offset = -1
                                              allocation_samples=signalSearchRange)

Distractor_Rep_Control_Signal = pnl.ControlSignal(modulates=[(pnl.SLOPE, Distractor_Rep)],
                                                  function=pnl.Linear,
                                                  variable=1.0,
                                                  cost_options=[pnl.CostFunctions.INTENSITY, pnl.CostFunctions.ADJUSTMENT],
                                                  intensity_cost_function=pnl.Exponential(scale=1, rate=1),
                                                  # adjustment_cost_function=pnl.Exponential(scale=1, rate=1, offset=-1),
                                                  #offset = -1
                                                  allocation_samples=signalSearchRange)

Umemoto_comp.add_model_based_optimizer(optimizer=pnl.OptimizationControlMechanism(
        agent_rep=Umemoto_comp,
        features=[Target_Stim.input_port,
                  Distractor_Stim.input_port,
                  Reward.input_port],
        feature_function=pnl.AdaptiveIntegrator(rate=1.0),
        objective_mechanism=pnl.ObjectiveMechanism(
                monitor_for_control=[Reward,
                                     (Decision.output_ports[pnl.PROBABILITY_UPPER_THRESHOLD], 1, -1)],
        ),
        function=pnl.GridSearch(save_values=True),
        control_signals=[Target_Rep_Control_Signal,
                         Distractor_Rep_Control_Signal],
        compute_reconfiguration_cost=pnl.Distance(
                metric=pnl.EUCLIDEAN)  # 0.8046),
))
Umemoto_comp.enable_model_based_optimizer = True
# Umemoto_comp.model_based_optimizer.set_log_conditions('value')



print(Target_Rep.loggable_items)
nTrials = 3
targetFeatures = [w_t]
flankerFeatures_inc = [w_d]
reward = [0]#[100]

targetInputList = targetFeatures
flankerInputList = flankerFeatures_inc
rewardList = reward

stim_list_dict = {
    Target_Stim: targetInputList,
    Distractor_Stim: flankerInputList,
     Reward: rewardList
}

Umemoto_comp.run(num_trials=nTrials,
                 inputs=stim_list_dict)


# print("\n\n---------  DISTRACTOR REP  ---------")
# Distractor_Rep.log.print_entries()
# print("\n\n---------  TARGET REP  ---------")
# Target_Rep.log.print_entries()
# print("\n\n---------  AUTOMATIC COMPONENT  ---------")
# Automatic_Component.log.print_entries()
# print("\n\n---------  DECISION  ---------")
# Decision.log.print_entries()
#
# print("\n\n---------  MODEL BASED OPTIMIZER  ---------")
# Umemoto_comp.model_based_optimizer.log.print_entries()
# print("\n\n---------  Target_Stim  ---------")
#
# Target_Stim.log.print_entries()

print(Umemoto_comp.model_based_optimizer.saved_samples)
print(Umemoto_comp.model_based_optimizer.saved_values)



"""

bistable percepts

"""



import numpy as np
import psyneulink as pnl
import pytest
from itertools import product


expected_3_10 = [[ 205.67990124], [ 205.536034],   [ 206.29612605],
                 [-204.87230198], [-204.98539771], [-205.35434273]]
expected_8_10 = [[-71427.62150144271], [-71428.44255569541],
                 [-71427.73782852193], [-71428.18340850921],
                 [-71428.10767225616], [-71428.22607075438],
                 [-71427.55903615047], [-71427.81981141337],
                  [67029.19595769834], [67028.98515147284],
                  [67029.00062228851], [67029.22270778783],
                  [67029.64637519913], [67028.31812638397],
                  [67028.98446253323], [67028.45363893337]]


@pytest.mark.model
@pytest.mark.benchmark
@pytest.mark.parametrize("n_nodes,n_time_steps,expected", [
    pytest.param(3, 10, expected_3_10, id="3-10"),
    pytest.param(8, 10, expected_8_10, id="8-10"),
])
@pytest.mark.parametrize("mode", [
    'Python',
    pytest.param('LLVM', marks=[pytest.mark.llvm]),
    pytest.param('LLVMExec', marks=[pytest.mark.llvm]),
    pytest.param('LLVMRun', marks=[pytest.mark.llvm]),
    pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
    pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda]),
])
def test_necker_cube(benchmark, mode, n_nodes, n_time_steps, expected):
    # this code only works for N_PERCEPTS == 2
    ALL_PERCEPTS = ['a', 'b']

    # variables
    n_nodes_per_percepts = n_nodes
    excit_level = 1
    inhib_level = 1
    node_dict = {percept: None for percept in ALL_PERCEPTS}

    def get_node(percept, node_id):
        """helper func for creating a node"""
        tm_function = pnl.Linear(slope=1, intercept=0)
        tm_integrator_mode = True
        tm_integration_rate = .5
        node_ = pnl.TransferMechanism(
            name='{percept}-{node_id}'.format(percept=percept, node_id=node_id),
            function=tm_function,
            integrator_mode=tm_integrator_mode,
            integration_rate=tm_integration_rate,
            default_variable=np.zeros((1,)),
        )
        return node_

    # init all nodes, save them in list and dict form
    for percept in ALL_PERCEPTS:
        node_dict[percept] = [
            get_node(percept, i) for i in range(n_nodes_per_percepts)
        ]

    # init composition
    bp_comp = pnl.Composition()

    # MODIFIED 4/4/20 OLD:  PASSES IN PYTHON, BUT NEEDS RESULTS B BELOW
    # within-percept excitation
    for percept in ALL_PERCEPTS:
        for node_i, node_j in product(node_dict[percept], node_dict[percept]):
            if node_i is not node_j:
                bp_comp.add_linear_processing_pathway(
                    pathway=(node_i, [excit_level], node_j))

    # inter-percepts inhibition
    for node_i, node_j in zip(node_dict[ALL_PERCEPTS[0]],
                              node_dict[ALL_PERCEPTS[1]]):
        bp_comp.add_linear_processing_pathway(
            pathway=(node_i, [-inhib_level], node_j))
        bp_comp.add_linear_processing_pathway(
            pathway=(node_j, [-inhib_level], node_i))

    # turn off report
    reportOutputPref = False

    # make sure all nodes are both input and outputs
    # # MODIFIED 4/25/20 OLD:
    # for node in bp_comp.nodes:
    #     bp_comp.add_required_node_role(node, pnl.NodeRole.INPUT)
    #     bp_comp.add_required_node_role(node, pnl.NodeRole.OUTPUT)
    # MODIFIED 4/25/20 NEW:
    for node in bp_comp.nodes:
        bp_comp.require_node_roles(node, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT])
    # MODIFIED 4/25/20 END
        # turn off report
        node.reportOutputPref = reportOutputPref

    # # MODIFIED 4/4/20 NEW:  [PASSES ALL TESTS, BUT NEEDS RSEULTS A BELOW]
    # # within-percept excitation
    # for percept in ALL_PERCEPTS:
    #     for node_i, node_j in product(node_dict[percept], node_dict[percept]):
    #         if node_i is not node_j:
    #             bp_comp.add_linear_processing_pathway(
    #                 pathway=((node_i, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT]), [excit_level], (node_j, [pnl.NodeRole.INPUT,
    #                                                                                                pnl.NodeRole.OUTPUT])))
    #
    # # inter-percepts inhibition
    # for node_i, node_j in zip(node_dict[ALL_PERCEPTS[0]],
    #                           node_dict[ALL_PERCEPTS[1]]):
    #     bp_comp.add_linear_processing_pathway(
    #         pathway=((node_i, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT]), [-inhib_level], (node_j, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT])))
    #     bp_comp.add_linear_processing_pathway(
    #         pathway=((node_j, [pnl.NodeRole.INPUT, pnl.NodeRole.OUTPUT]), [-inhib_level], (node_i, [pnl.NodeRole.INPUT,
    #                                                                                         pnl.NodeRole.OUTPUT])))

    # turn off report
    reportOutputPref = False
    # MODIFIED 4/4/20 END

    # bp_comp.show_graph()

    input_dict = {
        node_: np.random.normal(size=(n_time_steps,))
        for node_ in bp_comp.nodes
    }

    # run the model
    res = bp_comp.run(input_dict, num_trials=n_time_steps, bin_execute=mode)
    np.testing.assert_allclose(res, expected)

    # Test that order of CIM ports follows order of Nodes in self.nodes
    for i in range(n_nodes):
        a_name = "a-{}".format(i)
        assert a_name in bp_comp.input_CIM.input_ports.names[i]
        assert a_name in bp_comp.output_CIM.output_ports.names[i]
        b_name = "b-{}".format(i)
        assert b_name in bp_comp.input_CIM.input_ports.names[i + n_nodes]
        assert b_name in bp_comp.output_CIM.output_ports.names[i + n_nodes]

    if benchmark.enabled:
        benchmark.group = "Necker Cube {}-{}".format(n_nodes, n_time_steps)
        benchmark(bp_comp.run, input_dict, num_trials=n_time_steps, bin_execute=mode)


@pytest.mark.model
@pytest.mark.benchmark(group="Necker Cube")
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=[pytest.mark.llvm]),
                                  pytest.param('LLVMExec', marks=[pytest.mark.llvm]),
                                  pytest.param('LLVMRun', marks=[pytest.mark.llvm]),
                                  pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                  pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                  ])
def test_vectorized_necker_cube(benchmark, mode):

    Build_N_Matrix = np.zeros((16,5))
    Build_N_Matrix[0,:] = [0, 1, 3, 4, 8]
    Build_N_Matrix[1,:] = [1, 0, 2, 5, 9]
    Build_N_Matrix[2,:] = [2, 1, 3, 6, 10]
    Build_N_Matrix[3,:] = [3, 0, 2, 7, 11]
    Build_N_Matrix[4,:] = [4, 5, 7, 0, 12]
    Build_N_Matrix[5,:] = [5, 4, 6, 1, 13]
    Build_N_Matrix[6,:] = [6, 5, 7, 2, 14]
    Build_N_Matrix[7,:] = [7, 4, 6, 3, 15]
    Build_N_Matrix[8,:] = [8, 9, 11, 12, 0]
    Build_N_Matrix[9,:] = [9, 8, 10, 13, 1]
    Build_N_Matrix[10,:] = [10, 9, 11, 14, 2]
    Build_N_Matrix[11,:] = [11, 8, 10, 15, 3]
    Build_N_Matrix[12,:] = [12, 13, 15, 8, 4]
    Build_N_Matrix[13,:] = [13, 12, 14, 9, 5]
    Build_N_Matrix[14,:] = [14, 13, 15, 10, 6]
    Build_N_Matrix[15,:] = [15, 12, 14, 11, 7]

    Build_N_Matrix = Build_N_Matrix.astype(int)

    Necker_Matrix = np.zeros((16,16))
    Necker_Matrix = Necker_Matrix.astype(int)


    excite = 1
    inhibit = -2

    for x in range(0,16):
        Necker_Matrix[Build_N_Matrix[x,0], Build_N_Matrix[x,1]] = excite
        Necker_Matrix[Build_N_Matrix[x,0], Build_N_Matrix[x,2]] = excite
        Necker_Matrix[Build_N_Matrix[x,0], Build_N_Matrix[x,3]] = excite
        Necker_Matrix[Build_N_Matrix[x,0], Build_N_Matrix[x,4]] = inhibit

    comp2 = pnl.Composition()

    node3 = pnl.TransferMechanism(
        name='node3',
        function=pnl.Linear(slope = 1, intercept = 0),
        integrator_mode = True,
        integration_rate = .5,
        default_variable=np.zeros((1,16)),
    )


    #integrator function ((1-rate)*previous_value + rate*current_input) * mechanism_function

    node4 = pnl.TransferMechanism(
        name='node4',
        function=pnl.Linear(slope = 1, intercept = 0),
        integrator_mode = True,
        integration_rate = .5,
        default_variable=np.zeros((1,16)),
    )


    connect_3_4 = Necker_Matrix
    connect_4_3 = Necker_Matrix

    weights_3_4 = pnl.MappingProjection(
        name='connect_3_4',
        matrix=connect_3_4,
    )

    weights_4_3 = pnl.MappingProjection(
        name='connect_4_3',
        matrix=connect_3_4,
    )

    comp2.add_linear_processing_pathway(pathway = (node3, connect_3_4, node4, connect_4_3, node3))
    input_dict = {node3: np.random.random((1,16)),
                  node4: np.random.random((1,16))
                 }

    result = comp2.run(input_dict, num_trials=10, bin_execute=mode)
    assert np.allclose(result,
            [[ 2636.29181172,  -662.53579899,  2637.35386946,  -620.15550833,
               -595.55319772,  2616.74310649,  -442.74286574,  2588.4778162 ,
                725.33941441, -2645.25148476,   570.96811513, -2616.80319979,
              -2596.82097419,   547.30466563, -2597.99430789,   501.50648114],
             [ -733.2213593 ,  2638.81033464,  -578.76439993,  2610.55912376,
               2590.69244696,  -555.19824432,  2591.63200098,  -509.58072358,
              -2618.88711219,   682.65814776, -2620.18294962,   640.09719335,
                615.39758884, -2599.45663784,   462.67291695, -2570.99427346]])

    if benchmark.enabled:
        benchmark(comp2.run, input_dict, num_trials=10, bin_execute=mode)



from . import components
from . import compositions
from . import globals
from . import llvm
from . import scheduling
from . import rpc

from .components import *
from .compositions import *
from .globals import *
from .llvm import *
from .scheduling import *
from .rpc import *

__all__ = list(components.__all__)
__all__.extend(llvm.__all__)
__all__.extend(compositions.__all__)
__all__.extend(globals.__all__)
__all__.extend(scheduling.__all__)
__all__.extend(rpc.__all__)
